#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "config.h"
#include "imapfilter.h"
#include "data.h"


extern char logfile[PATH_MAX];
extern unsigned int options;
extern unsigned int flags;
extern unsigned int interval;
extern long timeout;
extern char *home;

#ifdef ENCRYPTED_PASSWORDS
char *passphr;			/* Master password to access the passwords
				   file. */
#endif


/*
 * Find the path to configuration file, open it and call parse_config().
 */
int read_config(char *cfg)
{
    int r;
    FILE *fd;
    char *c = NULL;

    if (!cfg) {
	cfg = c = (char *)xmalloc(PATH_MAX * sizeof(char));
	snprintf(cfg, PATH_MAX, "%s/%s", home, ".imapfilterrc");
    }
#ifdef DEBUG
    printf("debug: configuration file: '%s'\n", cfg);
#endif
#ifdef CHECK_PERMISSIONS
    check_file_perms(cfg, S_IRUSR | S_IWUSR);
#endif
    fd = fopen(cfg, "r");

    if (!fd)
	fatal(ERROR_FILE_OPEN, "imapfilter: opening config file %s; %s\n",
	      cfg, strerror(errno));

    if (c)
	xfree(c);

    if ((r = parse_config(fd)))
	fatal(ERROR_CONFIG_PARSE,
	      "imapfilter: parse error in config file at row %d\n", r);

    fclose(fd);

#ifdef DEBUG
    printf("debug: options: %0#10x\n", options);
#endif

    return 0;
}


/*
 * Parse configuration file.
 */
int parse_config(FILE * fd)
{
    int i, r = 0;
    unsigned int row = 0;
    char line[LINE_MAX];
    regex_t creg[13];
    regmatch_t match[11];
    const char *reg[13] = {
	"^([[:blank:]]*\n|#.*\n)$",

	"^[[:blank:]]*ACCOUNT[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
	"(([[:graph:]]+):([[:graph:]]*)|([[:graph:]]+))@([[:alnum:].-]+)"
	"(:[[:digit:]]+)?[[:blank:]]*([[:blank:]]SSL|[[:blank:]]SSL2|"
	"[[:blank:]]SSL3|[[:blank:]]TLS1)?[[:blank:]]*\n$",

	"^[[:blank:]]*FOLDER[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
	"([[:print:]]+)[[:blank:]]*\n$",

	"^[[:blank:]]*FILTER[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]*"
	"([[:blank:]]OR|[[:blank:]]AND)?[[:blank:]]*\n$",

	"^[[:blank:]]*ACTION[[:blank:]]+(DELETE|"
	"COPY[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)|"
	"MOVE[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)|"
	"RCOPY[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
	"(\"[[:print:]]*\"|[[:graph:]]+)|"
	"RMOVE[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
	"(\"[[:print:]]*\"|[[:graph:]]+)|"
	"FLAG[[:blank:]]+(REPLACE|ADD|REMOVE)[[:blank:]]+"
	"([[:alpha:],]+)|"
	"LIST)[[:blank:]]*([[:graph:]]*)[[:blank:]]*\n$",

	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
	"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
	"(ANSWERED|DELETED|DRAFT|FLAGGED|NEW|OLD|RECENT|SEEN|UNANSWERED|"
	"UNDELETED|UNDRAFT|UNFLAGGED|UNSEEN)[[:blank:]]*\n$",

	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
	"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
	"(BCC|BODY|CC|FROM|SUBJECT|TEXT|TO)[[:blank:]]+(\"[[:print:]]*\"|"
	"[[:graph:]]+)[[:blank:]]*\n$",

	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
	"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
	"(HEADER)[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]+"
	"(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]*\n$",

	"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
	"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
	"(LARGER|SMALLER|OLDER|NEWER)[[:blank:]]+([[:digit:]]+)"
	"[[:blank:]]*\n$",

	"^[[:blank:]]*JOB[[:blank:]]+([[:alnum:],_-]+)[[:blank:]]+"
	"([[:alnum:],_-]+)[[:blank:]]*\n$",

	"^[[:blank:]]*(SET[[:blank:]])?[[:blank:]]*(LOGFILE)"
	"[[:blank:]]*=[[:blank:]]*([[:print:]]+)\n$",

	"^[[:blank:]]*(SET[[:blank:]])?[[:blank:]]*(EXPUNGE|HEADERS|NAMESPACE|"
	"SUBSCRIBE|WARNING)[[:blank:]]*=[[:blank:]]*(YES|NO)[[:blank:]]*\n$",

	"^[[:blank:]]*(SET[[:blank:]])?[[:blank:]]*(DAEMON|TIMEOUT)"
	"[[:blank:]]*=[[:blank:]]*([[:digit:]]+)\n$"
    };

    for (i = 0; i < 13; i++)
	regcomp(&creg[i], reg[i], REG_EXTENDED | REG_ICASE);

    /* First process all the variables.  This is done because some variables'
       arguments are used before the other commands are processed. */
    while (fgets(line, LINE_MAX - 1, fd))
	if (!regexec(&creg[10], line, 4, match, 0) ||
	    !regexec(&creg[11], line, 4, match, 0) ||
	    !regexec(&creg[12], line, 4, match, 0))
	    set_options(line, match);

    /* Then rewind and process everything else. */
    fseek(fd, 0L, SEEK_SET);
    while (fgets(line, LINE_MAX - 1, fd)) {
	row++;
	if (!regexec(&creg[0], line, 0, match, 0))
	    continue;
	else if (!regexec(&creg[1], line, 9, match, 0))
	    set_account(line, match);
	else if (!regexec(&creg[2], line, 3, match, 0))
	    r = set_mboxgrp(line, match);
	else if (!regexec(&creg[3], line, 3, match, 0))
	    r = set_filter(line, match);
	else if (!regexec(&creg[4], line, 11, match, 0))
	    r = set_action(line, match);
	else if (!regexec(&creg[5], line, 5, match, 0))
	    r = set_mask(line, match, MASK_MATCH_1);
	else if (!regexec(&creg[6], line, 6, match, 0))
	    r = set_mask(line, match, MASK_MATCH_2);
	else if (!regexec(&creg[7], line, 7, match, 0))
	    r = set_mask(line, match, MASK_MATCH_3);
	else if (!regexec(&creg[8], line, 6, match, 0))
	    r = set_mask(line, match, MASK_MATCH_4);
	else if (!regexec(&creg[9], line, 3, match, 0))
	    r = set_job(line, match);
	/* Skip variable processing. */
	else if (!regexec(&creg[10], line, 4, match, 0) ||
		 !regexec(&creg[11], line, 4, match, 0) ||
		 !regexec(&creg[12], line, 4, match, 0));
	else
	    return row;

	if (r == ERROR_CONFIG_PARSE)
	    return row;
    }

    for (i = 0; i < 13; i++)
	regfree(&creg[i]);

    destroy_data();

    return 0;
}


/*
 * Set other options found in config file.
 */
void set_options(char *line, regmatch_t * match)
{
    if (!strncasecmp(line + match[2].rm_so, "logfile", 7)) {
	if (!*logfile)
	    strncat(logfile, line + match[3].rm_so,
		    min((match[3].rm_eo - match[3].rm_so), PATH_MAX - 1));
    } else if (!strncasecmp(line + match[2].rm_so, "expunge", 7)) {
	if (!strncasecmp(line + match[3].rm_so, "yes", 3))
	    options |= OPTION_EXPUNGE;
	else
	    options &= ~(OPTION_EXPUNGE);
    } else if (!strncasecmp(line + match[2].rm_so, "header", 6)) {
	if (!strncasecmp(line + match[3].rm_so, "yes", 3))
	    options |= OPTION_HEADERS;
	else
	    options &= ~(OPTION_HEADERS);
    } else if (!strncasecmp(line + match[2].rm_so, "namespace", 9)) {
	if (!strncasecmp(line + match[3].rm_so, "yes", 3))
	    options |= OPTION_NAMESPACE;
	else
	    options &= ~(OPTION_NAMESPACE);
    } else if (!strncasecmp(line + match[2].rm_so, "subscribe", 9)) {
	if (!strncasecmp(line + match[3].rm_so, "yes", 3))
	    options |= OPTION_SUBSCRIBE;
	else
	    options &= ~(OPTION_SUBSCRIBE);
    } else if (!strncasecmp(line + match[2].rm_so, "warning", 7)) {
	if (!strncasecmp(line + match[3].rm_so, "yes", 3))
	    options |= OPTION_WARNING;
	else
	    options &= ~(OPTION_WARNING);
    } else if (!strncasecmp(line + match[2].rm_so, "timeout", 7)) {
	errno = 0;
	timeout = strtol(line + match[3].rm_so, NULL, 10);
	if (errno)
	    timeout = 0;
    } else if (!strncasecmp(line + match[2].rm_so, "daemon", 6) &&
	       !(options & OPTION_DAEMON_MODE)) {
	options |= OPTION_DAEMON_MODE;
	errno = 0;
	interval = strtoul(line + match[3].rm_so, NULL, 10);
	if (errno)
	    interval = 0;
    }
}


#ifdef ENCRYPTED_PASSWORDS
/*
 * Open password file and call parse_passwords().
 */
int read_passwords(void)
{
    FILE *fd;
    char pwfile[PATH_MAX];

    if (!(flags & FLAG_BLANK_PASSWORD))
	return ERROR_CONFIG_PARSE;

    passphr = (char *)smalloc(PASSPHRASE_LEN);
    *passphr = 0;

    snprintf(pwfile, PATH_MAX, "%s/%s", home, ".imapfilter/passwords");
#ifdef DEBUG
    printf("debug: passwords file: '%s'\n", pwfile);
#endif

    if (!exists_file(pwfile))
	return ERROR_FILE_OPEN;

#ifdef CHECK_PERMISSIONS
    check_file_perms(pwfile, S_IRUSR | S_IWUSR);
#endif

    fd = fopen(pwfile, "r");

    if (!fd)
	fatal(ERROR_FILE_OPEN, "imapfilter: opening passwords file %s; %s\n",
	      pwfile, strerror(errno));

    parse_passwords(fd);

    fclose(fd);

    return 0;
}


/*
 *  Parse unencrypted password file.
 */
int parse_passwords(FILE * fd)
{
    int t = 3;
    char *pe = NULL;
    char user[USERNAME_LEN], serv[SERVER_LEN];
    unsigned char *buf;
    char *c, *cp, *line;
    regex_t creg;
    regmatch_t match[4];
    const char *reg = "([[:alnum:].-]+) ([[:graph:]]+) ([[:graph:]]+)";
    int r;

    do {
	fseek(fd, 0L, SEEK_SET);
	printf("Enter master passphrase: ");
	get_password(passphr, PASSPHRASE_LEN);
    } while ((r = decrypt_passwords(&buf, fd)) && --t);

    if (!t)
	return ERROR_PASSPHRASE;

    c = cp = sstrdup(buf);

    regcomp(&creg, reg, REG_EXTENDED | REG_ICASE);

    while ((line = strsep(&c, "\n")) &&
	   !regexec(&creg, line, 4, match, 0)) {
	user[0] = serv[0] = 0;

	strncat(serv, line + match[1].rm_so,
		min(match[1].rm_eo - match[1].rm_so, SERVER_LEN - 1));
	strncat(user, line + match[2].rm_so,
		min(match[2].rm_eo - match[2].rm_so, USERNAME_LEN - 1));

	if ((pe = (char *)find_password(user, serv)))
	    strncat(pe, line + match[3].rm_so,
		    min(match[3].rm_eo - match[3].rm_so, PASSWORD_LEN - 1));
    }

    regfree(&creg);
    sfree(cp);
    sfree(buf);

    return 0;
}


/*
 * Store encrypted passwords to file.
 */
int store_passwords(account_t * accts[])
{
    char pwfile[PATH_MAX];
    FILE *fd;

    snprintf(pwfile, PATH_MAX, "%s/%s", home, ".imapfilter/passwords");

    create_file(pwfile, S_IRUSR | S_IWUSR);

    fd = fopen(pwfile, "w");

    if (!fd)
	fatal(ERROR_FILE_OPEN, "imapfilter: opening passwords file %s; %s\n",
	      pwfile, strerror(errno));

    encrypt_passwords(fd, accts);

    fclose(fd);

    return 0;
}

#endif


/*
 * Create $HOME/.imapfilter directory.
 */
int create_homedir(void)
{
    char *hdn = ".imapfilter";

    if (home)
	if (chdir(home))
	    error("imapfilter: could not change directory; %s\n",
		  strerror(errno));

    if (!exists_dir(hdn))
	if (mkdir(hdn, S_IRUSR | S_IWUSR | S_IXUSR))
	    error("imapfilter: could not create directory %s; %s\n", hdn,
		  strerror(errno));

    return 0;
}


/*
 * Check if a file exists.
 */
int exists_file(char *fname)
{
    struct stat fs;

    if (access(fname, F_OK))
	return 0;

    stat(fname, &fs);
    if (!S_ISREG(fs.st_mode)) {
	error("imapfilter: file %s not a regular file\n", fname);
	return ERROR_FILE_OPEN;
    }
    return 1;
}


/*
 * Check if a directory exists.
 */
int exists_dir(char *dname)
{
    struct stat ds;

    if (access(dname, F_OK))
	return 0;

    stat(dname, &ds);
    if (!S_ISDIR(ds.st_mode)) {
	error("imapfilter: file %s not a directory\n", dname);
	return ERROR_FILE_OPEN;
    }
    return 1;
}


/*
 * Create a file with the specified permissions.
 */
int create_file(char *fname, mode_t mode)
{
    int fd = 0;

    if (!exists_file(fname)) {
	fd = open(fname, O_CREAT | O_WRONLY | O_TRUNC, mode);
	if (fd == -1) {
	    error("imapfilter: could not create file %s; %s\n", fname,
		  strerror(errno));
	    return ERROR_FILE_OPEN;
	}
	close(fd);
    }
    return 0;
}


#ifdef CHECK_PERMISSIONS
/*
 * Check the permissions of a file.
 */
int check_file_perms(char *fname, mode_t mode)
{
    struct stat fs;

    if (stat(fname, &fs)) {
	error("imapfilter: getting file %s status; %s\n", fname,
	      strerror(errno));
	return ERROR_TRIVIAL;
    }
    if (!S_ISREG(fs.st_mode)) {
	error("imapfilter: file %s not a regular file\n", fname);
	return ERROR_TRIVIAL;
    }
    if ((fs.st_mode & 00777) != mode) {
	error("imapfilter: warning: improper file %s permissions\n"
	      "imapfilter: warning: file's mode should be %o not %o\n",
	      fname, mode, fs.st_mode & 00777);
	return ERROR_TRIVIAL;
    }
    return 0;
}


/*
 * Check the permissions of a directory.
 */
int check_dir_perms(char *dname, mode_t mode)
{
    struct stat ds;

    if (stat(dname, &ds)) {
	error("imapfilter: getting file %s status; %s\n", dname,
	      strerror(errno));
	return ERROR_TRIVIAL;
    }
    if (!S_ISDIR(ds.st_mode)) {
	error("imapfilter: file %s not a directory\n", dname);
	return ERROR_TRIVIAL;
    }
    if ((ds.st_mode & 00777) != mode) {
	error("imapfilter: warning: improper dir %s permissions\n"
	      "imapfilter: warning: file's mode should be %o not %o\n",
	      dname, mode, ds.st_mode & 00777);
	return ERROR_TRIVIAL;
    }
    return 0;
}

#endif
